#!/usr/bin/env python
from ml_engines.supervised.classfication import Classify_ML
from ml_engines.utils.ml_tools import get_dimensions
from sklearn.linear_model import LogisticRegression
from typing import List
from unittest import TestCase
import numpy as np
import pandas as pd
np.random.seed(1000)

# Test datasets
loan = pd.DataFrame( np.random.randint(0, 2, size=(1000, 3)), columns=['Maritial Status',  'Gender', 'Employemeent'])
loan['Income'] = np.random.randint(30000, 120000, size=(1000, 1))
loan['Age'] = np.random.randint(18, 75, size=(1000, 1))
loan['Educations'] = np.random.randint(0, 7, size=(1000, 1))
loan['Number of Pervious Loans'] = np.random.randint(1, 8, size=(1000, 1))
loan['Loan Status'] = np.random.randint(0, 2, size=(1000, 1))
loan.index = np.arange(1, len(loan) + 1)

feature_names: List[str] = loan.columns.tolist()
feature_matrix: List[str] = loan.columns.tolist()[0:-1]
clf = Classify_ML(clf=LogisticRegression(C=1e9), clf_name='logreg',
                  feature_matrix=loan[feature_matrix[0:2]],
                  response_vector=loan['Loan Status'],
                  binary_clf=True,
                  feature_enginering=True,
                  feature_names=feature_names)

class TestClassification(TestCase):

    def test_dataframe_test(self):
        """
        Initial tests
        :return:
        """
        dimension: tuple = get_dimensions(loan)
        return self.assertEqual((1000, 8), dimension)






#!/usr/bin/env python
import pandas as pd
import numpy as np
import re
import os
from sklearn.feature_extraction.text import CountVectorizer, TfidfTransformer
from sklearn.linear_model import LogisticRegression
from sklearn import metrics
from sklearn.model_selection import train_test_split, cross_val_score
from urllib import request
from scipy.sparse.csr import csr_matrix
from sklearn.metrics import jaccard_similarity_score

# Alias
from typing import Dict, List
from pandas.core.frame import DataFrame

# Exceptions
from pandas.errors import EmptyDataError
from urllib.error import HTTPError
np.random.seed(10000)


class SentimentAnalysis(object):
    """
    class SentimentAnalysis(object)
            SentimentAnalysis(text: str, ngrams: tuple, stop_words: str = 'english',  **kwargs)

    Function Description
    -----------
    This is an implementation of Sentiment Analysis class for DSC550 (Bellevue University) Week-4 Assignment. This class has the following methods:

    Private Functions
    -----------------
    - __load_corpus__: this function will download and parse the corpus from github
    - __load_classifier__: load the given classifier from the pickle file
    - __load_file__: this function will load the given pickle file
    - __save_classifiers__: save the classifier to local pickle file
    - __predict__: predict the given text based on the trained corpus

    Public Functions
    ----------------
    - to_csv: parse the string comma separated representation
    - predict_text: call the private function __predict__
    - fit: fit the corpus
    - get_labels: label the given text as positive or negative
    - get_accuracy: get the accuracy of the classifier
    - get_model_eval: get the score of the cross-validation from the given scoring method
    - to_csv: function to convert unstructured text into a comma separated form
    - get_roc_auc: calcluate the Receiver Operating Curve AUC

    Parameters
    ----------
    - texts: given a string of text
    - ngrams: optional parameters with default value (1,3)
    - stop_words: optional parameters with default value english
    - kwargs: for now clf and evaluate_model are supported (i.e. {clf: MultiNomialNB(), evaluate_model: True})
        - clf_name: given the classifier name
        - clf: given the Scikit-Learn Classifier Object
        - evaluate_model: If set to True, then it will compute the accuracy using cross validations
        - scoring: given the scoring methods
        - save_clf: pickle the classifiers

    Examples
    --------
    >>> sentiment_analysis = SentimentAnalysis(texts=["There' s nothing special happening today."], **{'clf': logreg} )
    >>> sentiment_analysis.fit()
    >>> sentiment_analysis.get_labels()
    >>> {'Text': ["There' s nothing special happening today."], 'Label': 'Positive'}
    """

    def __init__(self, texts: List, ngrams: tuple = (1, 3), stop_words: str = 'english', **kwargs):
        self.texts: str = texts
        self.ngrams: tuple = ngrams
        self.stop_words: str = stop_words
        self.parameters: Dict = kwargs

        # Local variable
        self.labels_: Dict = {}
        self.probability_: Dict = {}
        self.corpus: DataFrame = pd.DataFrame()
        self.corpus_link: str = "https://raw.githubusercontent.com/ratmadjads/DSC550/master/datasets/googleplaystore_user_reviews.csv"
        self.clf = None
        self.y_pred_class = None
        self.y_pred_proba = None
        self.y_true = None
        self.accuracy: float = 0.0
        self.roc_auc_score: float = 0.0
        self.clf_pickle_path: str = "./clf_df.pickle"
        self.model_eval: float = 0.0
        self.clf_name: str = self.parameters.get('clf_name', 'Logistic Regression')  # Default is Logistic Regression
        self.file_name: str = re.sub(' ', '_', self.clf_name)
        self.tfidf_transform_path: str = "tfidf_transform.pickle"
        self.word_vector_path: str = "word_vect.pickle"
        self.train_dataset_path: str = "training.pickle"
        self.test_dataset_path: str = "testing.pickle"
        self.train_dtm_path: str = "trained_dtm.pickle"  # path to the trained document-term matrix
        self.test_dtm_path: str = "test_dtm.pickle"

        if not self.texts and not isinstance(texts, List):
            raise ValueError(f"ERROR: you must provide parameter texts")

    def to_csv(self, text: str) -> object:
        """
        Descriptions
        -------------
        Helper function to parse string comma separated representation
        
        Parameters
        ----------
        :text: given a non empty comma separated representation text
        :return: an object such as list with comma separte attributes or np.nan if the review field is empty
        """

        if not isinstance(text, str):
            raise TypeError(f"ERROR: {text} must be str not type of {type(text)}")

        # replace with np.nan if there is no review from the corpus
        if not '\"' in text:
            return np.nan

        else:

            # Get the first and last index of double quotes
            first_index: int = text.index("\"")
            last_index: int = text.rindex("\"")

            if first_index and last_index:
                review = text[first_index:last_index]
                row = text.replace(review, "").split(',')
                row[1] = review  # append review
                return row[0:3] + list(filter(lambda column: re.match(r"[0-9]+\.[0-9]+", column), row))

            return np.nan

    def __load_corpus__(self) -> DataFrame:
        """
        Description
        -----------
        Helper function to download sample corpus from Github and remove the Null value by the sentiment column.
        
        Parameters
        ----------
        :return: a corpus dataframe
        """
        try:
            response = request.urlopen(
                request.Request(self.corpus_link, headers={'User-Agent': 'DSC550-Week4 Assignment',
                                                           'Accept': "image/avif,image/webp,image/apng,image/*,*/*;q=0.8"
                                                           }))
        except HTTPError as e:
            raise HTTPError(f"ERROR: Unable to download corpus from {self.corpus_link}") from e

        raw_data: List[str] = response.read().decode('utf-8').split('\r\n')
        corpus_df: DataFrame = pd.DataFrame(
            pd.DataFrame(raw_data[1:], columns=['Texts'])['Texts'].astype(str).apply(self.to_csv).dropna().tolist(),
            columns=raw_data[0].split(','))
        corpus_df = corpus_df.dropna(subset=['Sentiment'])
        corpus_df = corpus_df.query("Sentiment != 'Neutral'")  # Select positive and negative comments
        corpus_df['Translated_Review'] = corpus_df['Translated_Review'].astype(str)
        self.corpus = corpus_df

        return self.corpus

    def __load_classifier__(self) -> DataFrame:
        """
        Description
        -----------
        Helper function to load pickle file from local directory
        
        Parameters
        ----------
        :return: an empty dataframe with default column Classifier_Name and Classifier or dataframe that contains trained classifiers
        """
        if not os.path.exists(self.clf_pickle_path):
            return pd.DataFrame(columns=['Classifier_Name', 'Classifier'])  # return empty dataframe

        clf_df: DataFrame = pd.read_pickle(self.clf_pickle_path)
        return clf_df

    def __save_classifiers__(self) -> bool:
        """
        Description
        -----------
        Helper function to save classifiers to the local directory
        
        Parameters
        ----------
        :return: a boolean value, if it is successful then return True, otherwise False
        """
        clf_df: DataFrame = self.__load_classifier__()
        if clf_df['Classifier_Name'].str.count(
                self.clf_name).sum() == 0:  # append if the classifier does not exists in the pickle file
            clf_df = clf_df.append([{'Classifier_Name': self.clf_name, 'Classifier': self.clf}], ignore_index=True)
            clf_df.index = np.arange(1, clf_df.shape[0] + 1)

        clf_df.to_pickle(f"{self.clf_pickle_path}")
        return clf_df.shape[0] > 0 and os.path.exists(self.clf_pickle_path)

    def __save_obj__(self, *args):
        """
        Description
        ------------
        Helper function to save training and testing dataset, as well as the document-term matricies
        
        Parameters
        ----------
        :kwargs: given an optional argumens
        :return:
        """
        transform_tfidf, word_vect, training, testing, train_df, test_df = args

        pd.DataFrame({'TFIDF': transform_tfidf}, index=[1]).to_pickle(f"./{self.file_name}_{self.tfidf_transform_path}")
        pd.DataFrame({'Word_Vect': word_vect}, index=[1]).to_pickle(f"./{self.file_name}_{self.word_vector_path}")
        pd.DataFrame({'Train_DTM': training}, index=[1]).to_pickle(f"./{self.file_name}_{self.train_dtm_path}")
        pd.DataFrame({'Test_DTM': testing}, index=[1]).to_pickle(f"./{self.file_name}_{self.test_dtm_path}")
        train_df.to_pickle(f"./{self.file_name}_{self.train_dataset_path}")
        test_df.to_pickle(f"./{self.file_name}_{self.test_dataset_path}")

    def __load__file__(self, file_path: str, column_name: str = "") -> object:
        """
        Description
        -----------
        Generic function to load all pickle files and return accordingly
        
        Parameters
        ----------
        :file_path: given a valid file path
        :column_name: given column name to be pickle
        :return: an object
        """
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"ERROR: Unable to load file {file_path} because it is missing")

        return pd.read_pickle(file_path)[column_name].iloc[0] if column_name else pd.read_pickle(file_path)

    def predict_text(self, text: str):
        """
        Description
        -----------
        Helper function to be called by the client program
        
        Parameters
        ----------
        text: given non empty text
        :return: None
        """
        if not text:
            raise ValueError("Text cannot be emtpy.")

        self.__predict__(text=text)

    def __predict__(self, transform_tfidf: object = None, word_vect: object = None, text: str = ""):
        """
        Description
        -----------
        This function will predict the given text or comment as positive or negative based on the training corpus
        
        Parameters
        ----------
        :transform_tfidf: an optional parameter for transform_tfidf
        :word_vect: an optional parameter for word_vect
        :text: an optional parameter for non empty text
        :return: None
        """
        if not self.clf:
            raise ValueError(f"Must call the fit function before making prediction")

        if not transform_tfidf and not word_vect:
            transform_tfidf = self.__load__file__(file_path=f"./{self.file_name}_{self.tfidf_transform_path}",
                                                  column_name='TFIDF')
            word_vect = self.__load__file__(file_path=f"./{self.file_name}_{self.word_vector_path}",
                                            column_name='Word_Vect')

        negative, positive = tuple(
            self.clf.predict_proba(transform_tfidf.transform(word_vect.transform([text] if text else self.texts)))[0])
        self.labels_ = {'Text': text if text else self.texts,
                        'Label': 'Positive' if negative < positive and positive > 0.7 else 'Negative'}
        self.probability_ = {'Positive': positive,
                             'Negative': negative}

        return None

    def fit(self) -> bool:
        """
        Description
        -----------
        Classify the text based on the corpus, the default classifier is Logistic Regression unless it is provided by the kwargs
        Note: The corpus is trained on Google Play Store Reviews dataset
        Parameters
        ----------
        :return: a boolean True if it is successful otherwise return False
        """

        # Load the classifier and predict the given text or comment, if the classifier exists in the pickle file
        clf_df: DataFrame = self.__load_classifier__()
        if not clf_df.empty and clf_df['Classifier_Name'].str.count(self.clf_name).sum() > 0:
            self.clf = clf_df.query(f"Classifier_Name == '{self.clf_name}' ").iloc[0][-1]  # get specific classifier
            self.__predict__()
            return self.clf != None

        corpus: DataFrame = self.__load_corpus__()  # Load Google Play Store Reviews
        if corpus.empty:
            raise EmptyDataError("Something went wrong, corpus should not be empty. Please re-run again")

        response_vector = corpus.Sentiment.map(dict(
            zip(list(corpus.Sentiment.unique()), [1, 0])))  # Label Positive comments as 1 and Negative comments as 0
        corpus['response_vector'] = response_vector

        # Split to training and testing data
        train_df, test_df = train_test_split(corpus[['Translated_Review', 'response_vector']],
                                             random_state=np.random.randint(100, 200), test_size=0.2)

        # create a collection of a matrix of token counts and transform the count matrix to a normalized tf-idf representation
        word_vect = CountVectorizer(ngram_range=self.ngrams, stop_words=self.stop_words)
        word_counts = word_vect.fit_transform(train_df['Translated_Review'])
        transform_tfidf = TfidfTransformer()
        training = transform_tfidf.fit_transform(word_counts)
        word_counts_vector = word_vect.transform(test_df['Translated_Review'])
        testing = transform_tfidf.transform(word_counts_vector)

        # train Google Play Store Reviews
        self.clf = self.parameters.get('clf') if self.parameters.get('clf', None) else LogisticRegression(C=1e35)
        self.clf.fit(training, train_df['response_vector'])
        self.y_pred_class = self.clf.predict(testing)
        self.y_pred_proba = self.clf.predict_proba(testing)[:, 1]

        # If kwargs evaluate_model is provided use cross-validation instead
        if self.parameters.get('evaluate_model'):
            try:
                self.model_eval = cross_val_score(self.clf, training, train_df['response_vector'],
                                                  scoring=self.parameters.get('scoring') if self.parameters.get(
                                                      'scoring') else 'accuracy', cv=10).mean()
                return self.model_eval

            except ValueError as e:
                scoring_methods: List[str] = metrics.SCORERS.keys()
                raise ValueError(f"ERROR: Use the following scoring methods {scoring_methods}") from e

        self.__predict__(transform_tfidf=transform_tfidf, word_vect=word_vect)
        self.accuracy = metrics.accuracy_score(y_true=test_df['response_vector'], y_pred=self.y_pred_class)
        self.roc_auc_score = metrics.roc_auc_score(y_true=test_df['response_vector'], y_score=self.y_pred_proba)
        self.y_true = test_df['response_vector']

        # If kwargs save_clf is provided saved the trainined classifier into pickle file
        if self.parameters.get('save_clf'):
            # Save the initial classifier for quicker access
            if clf_df.empty:
                save_classifiers: bool = self.__save_classifiers__()
                if not save_classifiers:
                    raise ValueError(f"Unable to pickle the given {self.clf_name} classifier")

                # initially save word_vect and tfidf_tranform
                self.__save_obj__(*(transform_tfidf, word_vect, training, testing, train_df, test_df))

            # Load, Lookup, and append
            else:
                if self.clf_name == 'Logistic Regression':
                    raise ValueError(f"{self.clf_name} is the default classifier, please provide different classifier.")

                save_classifiers: bool = self.__save_classifiers__()
                self.__save_obj__(*(transform_tfidf, word_vect, training, testing, train_df, test_df))
                if not save_classifiers:
                    raise ValueError(f"Unable to pickle the given {self.clf_name} classifier")

        return self.clf != None  # if the self.clf has not been instantiated it will return false

    def get_labels(self) -> Dict:
        """
        Description
        -----------
        Helper function to return labeled text as positive or negative
        
        Parameters
        ----------
        :return: a dictionary with labeled text
        """
        if not self.labels_:
            raise ValueError("ERROR: Please call the fit function or re-run without evaluate_model = False.")

        return self.labels_

    def get_accuracy(self) -> float:
        """
        Description
        -----------
        Helper function to return the accuracy of the given classifier
        
        Parameters
        -----------
        :return: the accuracy score in float
        """
        if not self.accuracy:
            if os.path.exists(f"./{self.file_name}_{self.train_dataset_path}") and os.path.exists(
                    f"./{self.file_name}_{self.test_dataset_path}") and os.path.exists(
                    f"./{self.file_name}_{self.test_dtm_path}"):
                test_df: DataFrame = self.__load__file__(file_path=f"./{self.file_name}_{self.test_dataset_path}")
                testing: csr_matrix = \
                self.__load__file__(file_path=f"./{self.file_name}_{self.test_dtm_path}")['Test_DTM'].iloc[0]
                self.y_pred_class = self.clf.predict(testing)
                return metrics.accuracy_score(y_true=test_df['response_vector'], y_pred=self.y_pred_class)

            raise ValueError("ERROR: Please call the fit function or re-run without evaluate_model = False.")

        return self.accuracy

    def get_roc_auc(self) -> float:
        """
        Description
        -----------
        Helper function to calculate the area under the curve
        
        Parameters
        ----------
        :return: area under the curve score
        """
        if not self.roc_auc_score:
            if os.path.exists(f"./{self.file_name}_{self.train_dataset_path}") and os.path.exists(
                    f"./{self.file_name}_{self.test_dataset_path}") and os.path.exists(
                    f"./{self.file_name}_{self.test_dtm_path}"):
                test_df: DataFrame = self.__load__file__(file_path=f"./{self.file_name}_{self.test_dataset_path}")
                testing: csr_matrix = \
                self.__load__file__(file_path=f"./{self.file_name}_{self.test_dtm_path}")['Test_DTM'].iloc[0]
                self.y_true = test_df['response_vector']
                self.y_pred_proba = self.clf.predict_proba(testing)[:, 1]
                return metrics.roc_auc_score(y_true=test_df['response_vector'], y_score=self.y_pred_proba)

        return self.roc_auc_score

    def get_model_eval(self, scoring: str = "") -> float:
        """
        Description
        -----------
         Helper function to return the score based on the scoring method (e.g. accuracy, roc_auc)
        
         Parameters
        -----------
        :scoring: given a valid scoring methods
        :return: model evaluation score
        """
        if not self.model_eval:
            feature_matrix_path: str = f"{self.file_name}_{self.train_dtm_path}"
            response_vector_path: str = f"{self.file_name}_{self.train_dataset_path}"

            training: csr_matrix = self.__load__file__(file_path=feature_matrix_path)['Train_DTM'].iloc[0]
            train_df: DataFrame = self.__load__file__(file_path=response_vector_path)
            if os.path.exists(feature_matrix_path) and os.path.exists(response_vector_path):
                return cross_val_score(self.clf, training, train_df['response_vector'],
                                       scoring=self.parameters.get('scoring') if self.parameters.get(
                                           'scoring') else scoring, cv=10).mean()

            raise ValueError("ERROR: Please call the fit function or re-run without evaluate_model = False.")

        return self.model_eval
    def word_similarity(word_one: str,  word_two: str) -> float:
        """
        Description
        -----------
        Helper function to calculate the jaccard similarity between 2 words, by converting text-documents into a matrix of token counts
        
        Parameters
        ----------
        :word_one: given a non-emtpy string 
        :word_two: given a non-emtpy string
        :return: a jaccard score between two strings
        
        Examples
        --------
        >>> words: List[str] = ['IS is waging war in Lybia', 'ISIL is waging war in Iraq']
        >>> word_similarity('IS is waging war in Lybia', 'ISIL is waging war in Iraq')
        >>> 0.42857142857142855
        """
        if not word_one and word_two:
            raise ValueError("word_one and word_two are required parameters.")
        
        words: List[str] = [word_one.lower(), word_two.lower()]
        vect = CountVectorizer() 
        dtm = vect.fit_transform(words) 
        counts: DataFrame = pd.DataFrame( dtm.toarray() , columns=vect.get_feature_names())
        
        return jaccard_similarity_score(counts.iloc[0].tolist() , counts.iloc[-1].tolist())



#!/usr/bin/env python
"""
Name: Rio Atmadja
Date: 23 May 2020
"""
import numpy as np
from typing import List
from pandas.core.frame import DataFrame
from pandas.core.series import Series
from botocore.exceptions import MissingParametersError
from typing import Dict
from numpy import ndarray
from sklearn.model_selection import train_test_split
from sklearn.pipeline import make_pipeline
from sklearn.model_selection import cross_val_score
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LogisticRegression
import pandas as pd
import random
import sklearn
random.seed(10000)

def get_random_features(feature_names: List[str]) -> List[str]:
    """
    This function will return random feature mtarix
    :param feature_names: given a list that contains the feature names
    :return: a list of feature matrix attributes
    """
    if not feature_names:
        raise ValueError("ERROR: feature names cannot be empty")

    lower_bound: int = np.random.randint(0, round(len(feature_names) + 1 / 2))
    upper_bound: int = np.random.randint(lower_bound, len(feature_names) + 1)
    return feature_names[lower_bound:upper_bound]

def get_best_features(df: DataFrame, *attrib) -> DataFrame:
    """
    This function will return the best features matrix
    :param df: given the dataframe
    :attrib: given a tuple of feature attributes
    :return: a dataframe
    """
    if df.empty:
        raise Warning("Must run clf.get_auto_accuracy first.")

    accuracy, sensitivity, specificity = attrib

    return df[(df['accuracy'] > accuracy) & (df['sensitivity'] > sensitivity) & (df['specificity'] > specificity)]


def auto_numeric_map(row: Series, col_names: List[str]) -> Series:
    """
    Convert string numeric into numbers
    :row: given a series of elements
    :col_names: given the colum names to be converted
    :return : numeric representation
    """
    row_data: List = row[col_names].tolist()
    results: List = []

    for data in row_data:
        if str(data).isdigit():
            results.append(int(data))

        elif len(str(row).split('.')) == 2 and str(row).split('.')[0].isdigit() and str(row).split('.')[-1].isdigit():
            results.append(float(data))

        else:
            results.append(pd.factorize(data)[0])

    return Series(results)

def save_columns(df: DataFrame) -> List[str]:
    """
    Helper function to save the
    :param df: given a data frame
    :return: a list that contains column names
    """
    return df.columns.tolist()


def get_possible_combinations(feature_names: List[str], iteration: int = int(np.random.randint(10, 50, size=1))) -> List[List]:
    """
    Helper function to remove duplicates feature matrix
    :feature_names: given the feature matrix
    :iteration: an optional parameter with a default iteration of 10 to 50 iterations
    :return: a list of list of feature matricies
    """
    if iteration <= 1:
        raise ValueError("ERROR: iteration must be greater than 1")

    if not feature_names:
        raise MissingParametersError(object_name="Required parameters", missing="feature_names")

    combinations: List[str] = sorted(set([','.join(get_random_features(feature_names=feature_names)) for n_features in range(iteration)]))

    return list(map(lambda row: row.split(','), filter(lambda row: len(row) > 1, map(lambda row: row, combinations))))

def train_test(feature_matrix: DataFrame, response_vector: Series, random_state: int = np.random.randint(100, 150)) -> Dict:
    """
    Helper function to create to split given feature matrix and response vectors into training and testing data
    :feature_matrix: given the feature matrix of mxn dimensions
    :response_vector: given a scalar vector
    :random_state: an optional param with random state from 100 to 150
    :return: a dictionary that contains testing and training data attributes
    """

    if not get_dimensions(feature_matrix) and not get_dimensions(response_vector):
        raise ValueError(
            f"ERROR Mismatch Dimensions Feature Matrix: {feature_matrix.shape} must equals Response Vector: {response_vector.shape}")

    X_train, X_test, y_train, y_test = train_test_split(feature_matrix, response_vector,
                                                        random_state=random_state)

    return {
        "X_train": X_train,
        "X_test": X_test,
        "y_train": y_train,
        "y_test": y_test

    }

def get_dimensions(vector) -> tuple:
    """
    Helper function to check the dimension of the given vectors and return the mxn dimension
    :vector: given mxn vector
    :return: a tuple of mxn object
    """
    if not type(vector) in [Series, DataFrame, ndarray]:
        raise TypeError("Error: Must be type of ndarray, Series, DataFrame")

    return vector.shape


def as_character(df: DataFrame, col_names: List[str]) -> DataFrame:
    """
    Helper function to convert dataframe of mxn into vector of characrters
    :param df: given an mxn data frame
    :param col_names: given column names to be converted
    :return: a set of mxn chars of data frame
    """
    if not col_names:
        raise MissingParametersError(object_name="Required parameters", missing="col_names")

    return df[col_names].apply(lambda col: pd.factorize(col)[0], axis=0)

def encode_catogries(df:DataFrame, col_names: List[str]) -> DataFrame:
    """
    Helper function to encode categories variables
    :df: given mxn dataframe
    :col_names: given the column names to be encoded
    :return: a dataframe
    """
    return pd.get_dummies( df[col_names], prefix=col_names)


def get_metric_scorers() -> List:
    """
    Helper functions to display all available scorers
    :return: a list of scorers attributes
    """
    return list(sklearn.metrics.SCORERS.keys())


def create_data_partition(clf, feature_matrix: DataFrame, response_vector: Series, scoring: str,
                          k_fold: int = random.randint(5, 9), scale: bool = False, mean: bool = False) -> float:
    """
    Helper function to perform cross validation, this function follow the convention of R createDataPartition functions.
    This function has a random default k_fold of 5 to 9 and if an optional parameter scale
    :clf: given ML algorithms
    :feature_matrix: given the feature matrix
    :response_vector: given the response vector
    :scoring: given the scoring method, please look at the function get_metric_scorers()
    :k_fold: an optional parameters, random default 5 to 9
    :scale: an optional parameters, default is False
    :mean: an optional parameters, default is False
    :return: a score based on the given method

    Examples:
    --------
    # >>> df = pd.DataFrame(np.random.randint(10,100,size=(100,10)), columns=[chr(random.randint(65,70) + random.randint(10,20)) for x in range(10)])
    # >>> df.index = np.arange(1,len(df) + 1 )
    # >>> df['A'] = np.random.randint(0,2, size=100)
    # >>> df.columns
    # >>> X = df[df.columns.tolist()[0:-1]]
    # >>> y = df['A']
    # >>> create_data_partition(clf=LogisticRegression(C=1e15), feature_matrix=X, response_vector=y, scoring='accuracy', k_fold=15, scale=True, mean=True)
    # >>> 0.4968253968253967

    """
    if not clf and not feature_matrix and not response_vector and not scoring:
        raise MissingParametersError(object_name="Required parameters", missing="clf, feature_matrix, response_vector, scoring")

    if scale:
        pipe = make_pipeline(StandardScaler(with_mean=mean), clf)
        return cross_val_score(pipe, feature_matrix, response_vector, cv=k_fold, scoring=scoring).mean()

    return cross_val_score(clf, feature_matrix, response_vector, cv=k_fold, scoring=scoring).mean()


def correlation(df: DataFrame) -> DataFrame:
    """
    Helper functions to calculate the correlations between variables in a given dataframe
    :df: given a mxn matricies
    :return: a dataframe with correlation attributes

    | Examples
    |--------
    | >>> df = pd.DataFrame({'Soldiers Casualties': np.random.randint(10,15, size=(1092,)),
    |                        'Unknown Fighters': np.random.randint(10,12, size=(1092,)) ,
    |                        'Civilian Casualties': np.random.randint(10,15, size=(1092,)),
    |                        'Local Militia': np.random.randint(10,15, size=(1092,)) ,
    |                        'Air Strikes': np.random.randint(10,15, size=(1092,))},
    |                         pd.date_range(start="2015-01-04", end="2017-12-30", freq='1D'))
    | >>> correlation(df=df)
    | >>>
    """
    if df.empty:
        raise ValueError("ERROR: Data Frame cannot be emtpy")

    if not isinstance(df, DataFrame):
        raise TypeError(f"ERROR: Must be type of pandas.core.frame.DataFrame not type {type(df)}")

    return pd.DataFrame({row: [sum((df[row] - df[row].mean()) * (df[col] - df[col].mean())) / (
                (df.shape[0] - 1) * df[row].std() * df[col].std()) for col in df.columns] for row in df.columns},
                        index=df.columns)

def covariance(df: DataFrame) -> DataFrame:
    """
    Helper functions to calculate the covariance between variables in a given dataframe
    :df: given a mxn matricies
    :return: a dataframe with correlation attributes

    |Examples
    |--------
    | >>> df = pd.DataFrame({'Soldiers Casualties': np.random.randint(10,15, size=(1092,)),
    |                        'Unknown Fighters': np.random.randint(10,12, size=(1092,)) ,
    |                        'Civilian Casualties': np.random.randint(10,15, size=(1092,)),
    |                        'Local Militia': np.random.randint(10,15, size=(1092,)) ,
    |                        'Air Strikes': np.random.randint(10,15, size=(1092,))},
    |                         pd.date_range(start="2015-01-04", end="2017-12-30", freq='1D'))
    | >>> covariance(df=df)
    | >>>
    """
    if df.empty:
        raise ValueError("ERROR: Data Frame cannot be emtpy")

    if not isinstance(df, DataFrame):
        raise TypeError(f"ERROR: Must be type of pandas.core.frame.DataFrame not type {type(df)}")

    return pd.DataFrame(
        {row: [sum((df[row] - df[row].mean()) * (df[col] - df[col].mean())) / ((df.shape[0] - 1)) for col in df.columns]
         for row in df.columns}, index=df.columns)

def find_best_classification_parameters(df: DataFrame, response: str, iterations: int = 100, random_state: int = 150) -> DataFrame:
    """
    Helper function to find parameters with the highest accuracy and auc scores
    :df: given dataframe with mxn size
    :response: given the endogenous variable
    :iterations: number of iterations to generate random parameters, with a default of 100 iterations
    :return: a sorted list with the highest to lowest parameters values for further analysis
    """
    if not response and df.empty:
        raise ValueError("ERROR: df and response are required")

    results: List = []
    features: List[str] = df.columns.tolist()
    possible_parameters: List[str] = sorted(set(
        map(lambda _: ','.join(np.random.choice(features, np.random.randint(1, len(features)), replace=True).tolist()),
            range(iterations))))
    for parameter in possible_parameters:

        # Ignore the response variable
        if response not in parameter:
            # Split dataset into training and testing
            feature_columns: List[str] = parameter.split(',')
            X_train, X_test, y_train, y_test = train_test_split(df[feature_columns], df[response],
                                                                random_state=np.random.randint(random_state - 50,
                                                                                               random_state))

            model_accuracy: float = cross_val_score(LogisticRegression(C=1e25), X_train, y_train, scoring='accuracy',
                                                    cv=10).mean()
            model_auc: float = cross_val_score(LogisticRegression(C=1e25), X_train, y_train, scoring='roc_auc',
                                               cv=10).mean()

            results.append({'Parameters': parameter,
                            'Accuracy': model_accuracy,
                            'Null Accuracy': max(y_test.mean(), y_test.mean() - 1),
                            'AUC': model_auc
                            })

    return pd.DataFrame(results).sort_values(by=['Accuracy', 'AUC', 'Null Accuracy'], ascending=False)

def find_best_regression_parameters(clf, df: DataFrame, response: str, iterations: int = 100,
                                    random_state: int = 150) -> DataFrame:
    """
    Helper function to find parameters with the highest rmse, null rmse, and r-squared score
    :clf: given the classifer
    :df: given dataframe with mxn size
    :response: given the endogenous variable
    :iterations: number of iterations to generate random parameters, with a default of 100 iterations
    :return: a sorted list with the highest to lowest parameters values for further analysis
    """
    if not response and df.empty:
        raise ValueError("ERROR: df and response are required")

    results: List = []
    features: List[str] = df.columns.tolist()
    possible_parameters: List[str] = sorted(set(
        map(lambda _: ','.join(np.random.choice(features, np.random.randint(1, len(features)), replace=True).tolist()),
            range(iterations))))
    for parameter in possible_parameters:

        # Ignore the response variable
        if response not in parameter:
            # Split dataset into training and testing
            feature_columns: List[str] = parameter.split(',')
            X_train, X_test, y_train, y_test = train_test_split(df[feature_columns], df[response],
                                                                random_state=np.random.randint(random_state - 50,
                                                                                               random_state))

            model_rmse: float = cross_val_score(clf, X_train, y_train, scoring='neg_root_mean_squared_error',
                                                cv=10).mean()
            model_rsquared: float = cross_val_score(clf, X_train, y_train, scoring='r2', cv=10).mean()

            # Compute the null rmse
            y_null: ndarray = np.zeros_like(y_test, dtype=float)
            y_null.fill(y_test.mean())

            # Compute
            results.append({'Parameters': parameter,
                            'RMSE': -1 * cross_val_score(LinearRegression(), X_train, y_train,
                                                         scoring='neg_root_mean_squared_error', cv=10).mean(),
                            'Null RMSE': np.sqrt(metrics.mean_squared_error(y_test, y_null)),
                            'Rsquared': cross_val_score(LinearRegression(), X_train, y_train, scoring='r2', cv=10).mean(),
                            })

    return pd.DataFrame(results).sort_values(by=['RMSE', 'Null RMSE', 'Rsquared'], ascending=False)

def replace_missing_values(df: DataFrame) -> bool:
    """
    Helpler function to replace missing values with its median
    :df: given a dataframe with missing values
    :return:
    """
    missing_columns: List[str] = list(map(lambda missing_value: missing_value[0], filter(
        lambda missing_value: missing_value[0] if missing_value[-1] != 0 else None,
        df[df.columns[0:-1]].isna().sum().to_dict().items())))
    for column in missing_columns:
        if not re.sub(r'[0-9]+', '', str(df[column].dtypes)) in ['int', 'float']:
            raise TypeError("Must be type of integers or floats")

        df[column].fillna(df[column].median(), inplace=True)

    return sum(df.isna().sum().values) == 0

def tune_model_rmse(clf, df:DataFrame, feature_cols:List[str], response_col:str , scaled:bool=False, k_fold: int = 10, bagged: bool = False) -> float:
    """
    Helper function to calculate model rmse 
    :clf: given the classifier 
    :df: given the dataframe 
    :feature_cols: given a list of string of feature columns 
    :response_col: given the string representation of response vector 
    :scaled: optional args with default scaled equals False 
    :k_fold: optional args with default of k_fold equals to 10 
    :bagged: optional args if base estimator is provided it will compute the RMSE using the given classifier, otherwise it will use the default BaggingRegressor. 
    (Note: This flag is useful to reduce the variance of a machine learning method, especially for Decession Tree)
    :return:
    """
    if not clf and df.empty and not feature_cols and not response_col: 
        raise ValueError("ERROR: required parameters model_rmse(clf, df:DataFrame, feature_cols:List[str], response_col:str ) ")
    
    X_train,X_test,y_train,y_test = train_test_split(df[feature_cols], df[response_col], random_state=np.random.randint(100,200))
    if scaled: 
        return { 'Standardized RMSE': -1 * cross_val_score(make_pipeline(StandardScaler(), clf), X_train, y_train, cv=k_fold, scoring='neg_root_mean_squared_error' ).mean() } 
    
    if bagged: 
        return {'Bagged RMSE': -1 * cross_val_score(BaggingRegressor(base_estimator=clf if clf else None , oob_score=True, n_estimators=np.random.randint(100,500), bootstrap=True), X_train, y_train, cv=k_fold, scoring='neg_root_mean_squared_error' ).mean() }
    
    return { 'RMSE' : -1 * cross_val_score(clf, X_train, y_train, cv=k_fold, scoring='neg_root_mean_squared_error' ).mean() } 



#!/usr/bin/env python
import requests
import subprocess
from typing import Dict, List
from datetime import datetime
import os

class Kaggle(object):
    def __init__(self):
        self.time_stamp: str = datetime.utcnow().isoformat()

    def load_orion_pkg(self, pypi_pkg: str, linux_academy_server: str) -> bool:
        """
        This function will retrieve credentials from Linux Academy and download custom package from PyPI repository, everytime the notebook is starting
        :pypi_pkg: given the custom package from ORION project
        :linux_academy: given dynamic IP Address from Linux Academy
        :return: a boolean installation status
        """
        if not pypi_pkg:
            raise ValueError(f"Missing required parameter: {pypi_pkg}")

        kaggle_ip: str = requests.get('https://ifconfig.me').text  # Record Kaggle IP
        user_agent: Dict = {'User-Agent': f'(KAGGLE_REQUEST_{kaggle_ip}_{self.time_stamp}) '}  # Record request from Kaggle Notebook
        pypi_creds: str = requests.get(f'http://{linux_academy_server}/data', headers=user_agent).text

        # save output
        with open(os.path.join(os.path.abspath('.'), '.pypirc'), 'w') as f:
            f.write(pypi_creds)
        f.close()

        # Install orionmlengines
        response: List[str] = [resp.decode('utf-8') for resp in
                               subprocess.Popen(pypi_pkg, shell=True, stdout=subprocess.PIPE).stdout.readlines()]
        return response.__len__() != 0


#!/usr/bin/python
"""
Name: Rio Atmadja
Date: 11 May 2020
Description: Generic Classifier class
"""
import pandas as pd
import numpy as np
from sklearn.metrics import accuracy_score, roc_curve, confusion_matrix
from sklearn.model_selection import cross_val_score

# Aliases
from typing import List, Dict
from pandas.core.series import Series
from pandas.core.frame import DataFrame
from numpy import ndarray
from ml_engines.utils.constants import DEFAULT_CLF
from ml_engines.utils.ml_tools import get_random_features, train_test, as_character
np.random.seed(10000)

class Classify_ML(object):
    """
    This class is Generic helper classfiersn class
    """

    def __init__(self, clf, clf_name: str, feature_matrix: DataFrame, response_vector: Series, **param_args):
        """
        Once instaniated, this class will fit and split the training and testing data
        :clf: given the classfier
        :clf_name: given the classfier name
        :feature_matrix: given the independent variables
        :response_vector: given the dependent variable
        :param_args: given arbitrary arguments of kwargs optional parameters
        """

        # intantiates arguments
        self.param_args: Dict = param_args
        self.feature_matrix: DataFrame = feature_matrix
        self.response_vector: Series = response_vector
        self.clf = None
        self.feature_names_combinations: List[str] = []
        self.auto_accuracy: List[Dict] = []
        self.df = None
        self.clf_name = clf_name

        if not clf and not clf_name and not feature_matrix and not response_vector:
            raise AttributeError("ERROR: clf, clf_name, feature_matrix, response_vector are required.")

        if clf_name not in DEFAULT_CLF:
            raise ValueError(f"ERROR: available classfier for now {DEFAULT_CLF}")

        # split training and testing data here
        training_testing: Dict = train_test(feature_matrix=self.feature_matrix, response_vector=self.response_vector)

        self.X_train = training_testing.get("X_train")
        self.X_test = training_testing.get("X_test")
        self.y_train = training_testing.get("y_train")
        self.y_test = training_testing.get("y_test")

        self.clf = clf.fit(self.X_train, self.y_train)
        self.clf_original = clf

    def get_clf_name(self) -> str:
        """
        This function will return the classifier name
        :return: a string of classifier name
        """
        return self.clf_name

    def get_prediction_class(self) -> ndarray:
        """
        This function will return the prediction class from the given classfier
        :return: numpy array of the prediction class
        """
        return self.clf.predict(self.X_test)

    def get_predict_probability(self) -> ndarray:
        """
        This function will return all the prediction probabilites
        :return: numpy array of the prediction probabilities
        """
        return self.clf.predict_proba(self.X_test)

    def get_roc_auc_score(self, k_fold: int = np.random.randint(10, 15)) -> float:
        """
        This function calculate the roc_auc score, using cross validations technique.
        With a random kfolds from 10 to 15
        :k_fold: an optional k_fold parameter with an initial value of 10 - 15 k_fold
        :return: a roc_auc score in decimal points
        """
        if not self.is_binary():
            raise TypeError("True values must be in binary formats ")
        return cross_val_score(self.clf,
                               self.X_train,
                               self.y_train, cv=k_fold,
                               scoring="roc_auc").mean()  # Random K-folds 10-15

    def get_prediction_accuracy(self) -> float:
        """
        This function will return the prediction accuracy in decimal points
        :return the prediction accuracy in decimal points
        """
        return accuracy_score(self.get_prediction_class(), self.y_test)

    def get_null_accuracy(self) -> float:
        """
        This function will return the null accuracy from the true value vector
        :y_test: given the true value vector
        :return: the null accuracy from response vector
        """
        if self.param_args.get('binary_clf'):
            return max(self.y_test.mean(), 1 - self.y_test.mean())

        return float(self.y_test.value_counts().head(1) / len(self.y_test))

    def get_roc_curve(self) -> tuple:
        """
        This function will compute the ROC curve
        :return: a tuple that contains True postive, False postive rates and Thresholds
        """

        if not self.is_binary():
            raise TypeError("True values must be in binary formats ")

        fpr, tpr, threshold = roc_curve(self.y_test, self.get_predict_probability()[:, 1])
        return (fpr, tpr, threshold)

    def get_auto_accuracy(self, iteration: int = 1000) -> DataFrame:
        """
        This function will attempt to automatically tune feature matrix and find the best accuracy
        :iteration: an optional parameter with default iteration of 1000
        :return: pandas dataframe
        """

        feature_names: List[str] = self.param_args.get('feature_names')
        for i in range(iteration):
            self.feature_names_combinations.append(get_random_features(feature_enginering=self.param_args.get('feature_enginering') ,
                                                                       feature_names=feature_names))

        for feature in self.unique_feature_names():
            training_testing: Dict = train_test(feature_matrix=self.feature_matrix,
                                                response_vector=self.response_vector)

            X_train = training_testing.get("X_train")
            X_test = training_testing.get("X_test")
            y_train = training_testing.get("y_train")
            y_test = training_testing.get("y_test")

            y_pred_class = self.clf_original.fit(X_train, y_train).predict(X_test)
            results: Dict = self.calculate_cm(cm=confusion_matrix(y_test, y_pred_class), is_cm=True)
            results['features'] = feature
            self.auto_accuracy.append(results)

        self.df = pd.DataFrame.from_dict(self.auto_accuracy)
        return self.df

    def unique_feature_names(self) -> List[List]:
        """
        This function will return unique feature names
        :return: a list of list of unique feature names
        """
        if not self.feature_names_combinations:
            return []

        combination_feature_names: List[str] = sorted(set(list(filter(lambda column_name:
                                                                      column_name if len(column_name) else None,
                                                                      map(lambda features: ','.join(sorted(features)),
                                                                          self.feature_names_combinations)
                                                                      )
                                                               )
                                                          )
                                                      )

        return list(map(lambda feature: feature.split(','), combination_feature_names))

    def is_binary(self) -> bool:
        """
        Helpler function to check, if the true values contains binary response
        :return: a boolean value
        """
        return [0, 1] == sorted(pd.unique(self.y_test.tolist()).tolist())

    def calculate_cm(self, cm: ndarray = None, is_cm: bool = False) -> Dict:
        """

        This function will calculate the following confusion matrix attributes:
        - True Positive
        - True Negative
        - False Positive
        - False Negative
        - Sensitivity
        - Specificity
        - Accuracy
        - Error Rate
        - False Positive Rate
        :cm: an optional confusion matrix parameter, default is None
        :is_cm: an optional flag, set to True. If confusion matrix is provided
        :return: a dictionary of confusion matrix attributes
        """

        if not is_cm:
            cm = confusion_matrix(self.y_test, self.get_prediction_class())
        tp = cm[1][1]
        tn = cm[0][0]
        fp = cm[0][1]
        fn = cm[1][0]

        total: int = tp + tn + fp + fn
        actual_no: int = tn + fp
        actual_yes: int = fn + fp

        return {
            "accuracy": (tp + tn) / total,
            "error_rate": (fp + fn) / total,
            "sensitivity": tp / float(actual_yes),
            "specificity": tn / float(actual_no),
            "false_positve_rate": fp / float(actual_no)
        }



#!/usr/bin/env python
"""
Name: Rio Atmadja
Date: 11 May 2020
Description: Contasts variables for this project
"""
from typing import List, Dict
DEFAULT_CLF: List[str] = ['logreg', 'knn', 'dtree']
DEFAULT_RGR: List[str] = ['linreg']
MONTHS: Dict = {
    "January":"01",
    "February":"02",
    "March":"03",
    "April":"04",
    "May":"05",
    "June":"06",
    "July":"07",
    "August":"08",
    "September":"09",
    "October":"10",
    "November":"11",
    "December":"12"
}


#!/usr/bin/env python
"""
Name: Rio Atmadja
Date: 26 May 2020
Description: Helper class for orion-ml-engines
"""
import re
from ml_engines.utils.constants import MONTHS
def as_date(date_fmt: str) -> str:
    """
    Helper function to format date in YYYY-MM-DD
    :date_fmt: given the date format
    :return: a string in YYYY-MM-DD format
    """

    validate_date: Match = re.search(r"[a-z]{1,} [0-9]{1,2},\ {1,}[0-9]{4}",
                                     date_fmt.lower())  # validate date date format
    if not validate_date:
        raise ValueError("ERROR: Must be in format MM DD, YYYY")

    fmt_dates, fmt_year = tuple(date_fmt.split(','))
    month, day = tuple(fmt_dates.split(' '))
    day = "0" + day if int(day) < 10 else day

    return f"{fmt_year}-{MONTHS.get(month)}-{day}".strip()